@@ -26,7 +26,7 @@
 
 GRUB_MOD_LICENSE ("GPLv3+");
 
-static char *grub_datetime_names[] =
+static const char *grub_datetime_names[] =
 {
   "YEAR",
   "MONTH",
@@ -37,7 +37,7 @@ static char *grub_datetime_names[] =
   "WEEKDAY",
 };
 
-static char *
+static const char *
 grub_read_hook_datetime (struct grub_env_var *var,
                          const char *val __attribute__ ((unused)))
 {
@@ -50,7 +50,7 @@ grub_read_hook_datetime (struct grub_env_var *var,
       int i;
 
       for (i = 0; i < 7; i++)
-        if (! grub_strcmp (var->name, grub_datetime_names[i]))
+        if (grub_strcmp (var->name, grub_datetime_names[i]) == 0)
           {
             int n;
 
